from __future__ import annotations

import contextlib
import datetime
import os
import pathlib
from typing import Any

import pytest
import sqlalchemy as sa

from cacholote import config

does_not_raise = contextlib.nullcontext


def test_change_sessionmaker(tmp_path: pathlib.Path) -> None:
    old_sessionmaker = config.get().instantiated_sessionmaker
    new_db = "sqlite:///" + str(tmp_path / "dummy.db")

    with config.set(cache_db_urlpath=new_db):
        new_sessionmaker = config.get().instantiated_sessionmaker
        assert new_sessionmaker is not old_sessionmaker
    assert config.get().instantiated_sessionmaker is old_sessionmaker

    with config.set(sessionmaker=new_sessionmaker):
        assert config.get().instantiated_sessionmaker is new_sessionmaker
    assert config.get().instantiated_sessionmaker is old_sessionmaker

    config.set(cache_db_urlpath=new_db)
    assert config.get().instantiated_sessionmaker is new_sessionmaker

    config.set(sessionmaker=old_sessionmaker)
    assert config.get().instantiated_sessionmaker is old_sessionmaker


def test_change_cache_db_urlpath(tmp_path: pathlib.Path) -> None:
    old_db = config.get().cache_db_urlpath
    new_db = "sqlite:///" + str(tmp_path / "dummy.db")

    with config.set(cache_db_urlpath=new_db):
        assert str(config.get().engine.url) == config.get().cache_db_urlpath == new_db
    assert str(config.get().engine.url) == config.get().cache_db_urlpath == old_db

    config.set(cache_db_urlpath=new_db)
    assert str(config.get().engine.url) == config.get().cache_db_urlpath == new_db


@pytest.mark.parametrize(
    "key, reset",
    [
        ("cache_db_urlpath", True),
        ("create_engine_kwargs", True),
        ("cache_files_urlpath", False),
    ],
)
def test_set_engine_and_sessionmaker(
    tmp_path: pathlib.Path, key: str, reset: bool
) -> None:
    old_engine = config.get().engine
    old_sessionmaker = config.get().instantiated_sessionmaker

    kwargs: dict[str, Any] = {}
    if key == "cache_db_urlpath":
        kwargs[key] = "sqlite:///" + str(tmp_path / "dummy.db")
    elif key == "create_engine_kwargs":
        kwargs[key] = {"pool_recycle": 60}
    elif key == "cache_files_urlpath":
        kwargs[key] = str(tmp_path / "dummy_files")
    else:
        raise ValueError

    with config.set(**kwargs):
        if reset:
            assert config.get().engine is not old_engine
            assert config.get().instantiated_sessionmaker is not old_sessionmaker
        else:
            assert config.get().engine is old_engine
            assert config.get().instantiated_sessionmaker is old_sessionmaker
    assert config.get().engine is old_engine
    assert config.get().instantiated_sessionmaker is old_sessionmaker

    config.set(**kwargs)
    if reset:
        assert config.get().engine is not old_engine
        assert config.get().instantiated_sessionmaker is not old_sessionmaker
    else:
        assert config.get().engine is old_engine
        assert config.get().instantiated_sessionmaker is old_sessionmaker


def test_env_variables(tmp_path: pathlib.Path) -> None:
    # env variables
    old_environ = dict(os.environ)
    os.environ["CACHOLOTE_CACHE_DB_URLPATH"] = "sqlite://"

    # env file
    dotenv_path = tmp_path / ".env.cacholote"
    with dotenv_path.open("w") as f:
        f.write("CACHOLOTE_IO_DELETE_ORIGINAL=TRUE")

    config.reset(str(dotenv_path))
    try:
        assert config.get().cache_db_urlpath == "sqlite://"
        assert str(config.get().engine.url) == "sqlite://"
        assert config.get().io_delete_original is True
        assert str(config.get().engine.url) == "sqlite://"
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


@pytest.mark.parametrize("poolclass", ("NullPool", sa.pool.NullPool))
def test_set_poolclass(poolclass: str | sa.pool.Pool) -> None:
    config.set(create_engine_kwargs={"poolclass": poolclass})
    settings = config.get()
    assert settings.create_engine_kwargs["poolclass"] == sa.pool.NullPool
    assert isinstance(settings.engine.pool, sa.pool.NullPool)


@pytest.mark.parametrize(
    "expiration,raises",
    [
        (
            datetime.datetime.now(),
            pytest.raises(ValueError, match="Expiration is missing the timezone info."),
        ),
        (
            datetime.datetime.now().isoformat(),
            pytest.raises(ValueError, match="Expiration is missing the timezone info."),
        ),
        (datetime.datetime.now(tz=datetime.timezone.utc), does_not_raise()),
        (datetime.datetime.now(tz=datetime.timezone.utc).isoformat(), does_not_raise()),
    ],
)
def test_set_expiration(
    expiration: datetime.datetime | str,
    raises: contextlib.nullcontext,  # type: ignore[type-arg]
) -> None:
    with raises:
        config.set(expiration=expiration)


def test_create_engine_dict_kwargs() -> None:
    old_session_maker = config.get().instantiated_sessionmaker
    config.set(create_engine_kwargs={"connect_args": {"timeout": 30}})
    assert config.get().instantiated_sessionmaker is not old_session_maker



from __future__ import annotations

import contextlib
import hashlib
import importlib
import io
import pathlib
import subprocess
from typing import Any

import fsspec
import pytest
import pytest_httpserver
import pytest_structlog
import structlog

from cacholote import cache, config, decode, encode, extra_encoders, utils


@cache.cacheable
def cached_open(*args: Any, **kwargs: Any) -> fsspec.spec.AbstractBufferedFile:
    with fsspec.open(*args, **kwargs) as f:
        return f


@pytest.mark.parametrize("io_delete_original", [True, False])
def test_dictify_io_object(tmp_path: pathlib.Path, io_delete_original: bool) -> None:
    # Define readonly dir
    readonly_dir = str(tmp_path / "readonly")
    fsspec.filesystem("file").mkdir(readonly_dir)
    config.set(
        io_delete_original=io_delete_original, cache_files_urlpath_readonly=readonly_dir
    )

    # Create file
    tmpfile = tmp_path / "test.txt"
    fsspec.filesystem("file").pipe_file(tmpfile, b"test")
    tmp_hash = f"{fsspec.filesystem('file').checksum(tmpfile):x}"

    # Check dict and cached file
    actual = extra_encoders.dictify_io_object(open(tmpfile, "rb"))
    href = f"{readonly_dir}/{tmp_hash}.txt"
    local_path = f"{tmp_path}/cache_files/{tmp_hash}.txt"
    expected = {
        "type": "python_call",
        "callable": "cacholote.extra_encoders:decode_io_object",
        "args": (
            {
                "type": "text/plain",
                "href": href,
                "file:checksum": f"{fsspec.filesystem('file').checksum(local_path):x}",
                "file:size": 4,
                "file:local_path": local_path,
            },
            {},
        ),
        "kwargs": {"mode": "rb"},
    }
    assert actual == expected
    assert fsspec.filesystem("file").exists(tmpfile) is not io_delete_original

    # Use href when local_path is missing or corrupted
    fsspec.filesystem("file").mv(local_path, href)
    assert decode.loads(encode.dumps(actual)).read() == b"test"


@pytest.mark.parametrize("obj", [io.BytesIO(b"test"), io.StringIO("test")])
def test_dictify_bytes_io_object(
    tmp_path: pathlib.Path, obj: io.BytesIO | io.StringIO
) -> None:
    actual = extra_encoders.dictify_io_object(obj)["args"]
    obj_hash = hashlib.md5(f"{hash(obj)}".encode()).hexdigest()
    local_path = f"{tmp_path}/cache_files/{obj_hash}"
    type = (
        "text/plain"
        if importlib.util.find_spec("magic")
        else "application/octet-stream"
    )
    expected: tuple[dict[str, Any], ...] = (
        {
            "type": type,
            "href": local_path,
            "file:checksum": f"{fsspec.filesystem('file').checksum(local_path):x}",
            "file:size": 4,
            "file:local_path": local_path,
        },
        {},
    )
    assert actual == expected
    assert open(local_path).read() == "test"


@pytest.mark.parametrize("set_cache", ["file", "cads"], indirect=True)
def test_copy_from_http_to_cache(
    tmp_path: pathlib.Path,
    httpserver: pytest_httpserver.HTTPServer,
    set_cache: str,
) -> None:
    # cache-db to check
    con = config.get().engine.raw_connection()
    cur = con.cursor()

    # http server
    httpserver.expect_request("/test").respond_with_data(b"test")
    url = httpserver.url_for("/test")
    cached_basename = f"{fsspec.filesystem('http').checksum(url):x}"

    # cache http file
    fs, dirname = utils.get_cache_files_fs_dirname()
    for expected_counter in (1, 2):
        result = cached_open(url)

        # Check hits
        cur.execute("SELECT counter FROM cache_entries", ())
        assert cur.fetchall() == [(expected_counter,)]

        # Check result
        assert result.read() == b"test"

        # Check file in cache
        assert result.path == f"{dirname}/{cached_basename}"
        assert fs.exists(result.path)


def test_io_corrupted_files(
    tmp_path: pathlib.Path, httpserver: pytest_httpserver.HTTPServer
) -> None:
    # http server
    httpserver.expect_request("/test").respond_with_data(b"test")
    url = httpserver.url_for("/test")
    cached_basename = f"{fsspec.filesystem('http').checksum(url):x}"

    # cache file
    fs, dirname = utils.get_cache_files_fs_dirname()
    cached_open(url)

    # Warn if file is corrupted
    fs.touch(f"{dirname}/{cached_basename}")
    touched_info = fs.info(f"{dirname}/{cached_basename}")
    with pytest.warns(UserWarning, match="checksum mismatch"):
        result = cached_open(url)
    assert result.read() == b"test"
    assert fs.info(f"{dirname}/{cached_basename}") != touched_info

    # Warn if file is deleted
    fs.rm(f"{dirname}/{cached_basename}")
    with pytest.warns(UserWarning, match="No such file or directory"):
        result = cached_open(url)
    assert result.read() == b"test"
    assert fs.exists(f"{dirname}/{cached_basename}")


@pytest.mark.parametrize(
    "lock_timeout, raises_or_warns",
    (
        [None, pytest.warns(UserWarning, match="is locked")],
        [0, pytest.raises(TimeoutError, match="is locked")],
    ),
)
def test_io_locker(
    tmp_path: pathlib.Path,
    lock_timeout: float | None,
    raises_or_warns: contextlib.nullcontext,  # type: ignore[type-arg]
) -> None:
    config.set(lock_timeout=lock_timeout, raise_all_encoding_errors=True)
    # Create tmpfile
    tmpfile = tmp_path / "test.txt"
    fsspec.filesystem("file").touch(tmpfile)

    # Acquire lock
    fs, dirname = utils.get_cache_files_fs_dirname()
    file_path = f"{dirname}/{fsspec.filesystem('file').checksum(tmpfile):x}.txt"
    fs.touch(file_path)
    fs.touch(f"{file_path}.lock")

    process = subprocess.Popen(f"sleep 0.1; rm {file_path}.lock", shell=True)
    with raises_or_warns:
        cached_open(tmpfile)
    assert process.wait() == 0


@pytest.mark.parametrize("set_cache", ["cads"], indirect=True)
def test_content_type(tmp_path: pathlib.Path, set_cache: str) -> None:
    tmpfile = str(tmp_path / "test.grib")
    fsspec.filesystem("file").touch(tmpfile)
    fs, _ = utils.get_cache_files_fs_dirname()
    cached_grib = cached_open(tmpfile)
    assert fs.info(cached_grib)["ContentType"] == "application/x-grib"


@pytest.mark.parametrize("set_cache", ["cads"], indirect=True)
def test_io_logging(
    log: pytest_structlog.StructuredLogCapture, tmp_path: pathlib.Path
) -> None:
    config.set(logger=structlog.get_logger(), io_delete_original=True)

    # Cache file
    tmpfile = tmp_path / "test.txt"
    fsspec.filesystem("file").touch(tmpfile)
    cached_file = cached_open(tmpfile)

    tmp_urlpath = f"file://{tmpfile!s}"
    cached_urlpath = f"s3://{cached_file.path}"
    expected = [
        {
            "urlpath": cached_urlpath,
            "size": 0,
            "event": "start upload",
            "level": "info",
        },
        {
            "urlpath": cached_urlpath,
            "size": 0,
            "upload_time": log.events[1]["upload_time"],
            "event": "end upload",
            "level": "info",
        },
        {
            "urlpath": tmp_urlpath,
            "size": 0,
            "event": "start remove",
            "level": "info",
        },
        {
            "urlpath": tmp_urlpath,
            "size": 0,
            "remove_time": log.events[3]["remove_time"],
            "event": "end remove",
            "level": "info",
        },
        {
            "urlpath": cached_urlpath,
            "event": "retrieve cache file",
            "level": "info",
        },
    ]
    assert log.events == expected



from __future__ import annotations

import os
import pathlib
from collections.abc import Iterator
from typing import Any

import botocore.session
import psycopg
import pytest
import requests
from moto.moto_server.threaded_moto_server import ThreadedMotoServer

from cacholote import config, database


@pytest.fixture(scope="session")
def s3_server() -> Iterator[ThreadedMotoServer]:
    if "AWS_SECRET_ACCESS_KEY" not in os.environ:
        os.environ["AWS_SECRET_ACCESS_KEY"] = "foo"
    if "AWS_ACCESS_KEY_ID" not in os.environ:
        os.environ["AWS_ACCESS_KEY_ID"] = "foo"
    server = ThreadedMotoServer(ip_address="127.0.0.1", port=5555)
    server.start()
    yield server
    server.stop()


def create_test_bucket(
    server: ThreadedMotoServer, test_bucket_name: str
) -> dict[str, Any]:
    endpoint_url = f"http://{server._ip_address}:{server._port}/"
    client_kwargs = {"endpoint_url": endpoint_url}
    requests.post(f"{endpoint_url}moto-api/reset")
    session = botocore.session.Session()
    client = session.create_client("s3", **client_kwargs)
    client.create_bucket(Bucket=test_bucket_name)
    return client_kwargs


@pytest.fixture(autouse=True)
def set_cache(
    tmp_path: pathlib.Path,
    postgresql: psycopg.Connection[Any],
    request: pytest.FixtureRequest,
    s3_server: ThreadedMotoServer,
) -> Iterator[str]:
    param = getattr(request, "param", "file")
    if param.lower() == "cads":
        database._cached_sessionmaker.cache_clear()
        test_bucket_name = "test-bucket"
        client_kwargs = create_test_bucket(s3_server, test_bucket_name)
        with config.set(
            cache_db_urlpath=(
                f"postgresql+psycopg2://{postgresql.info.user}:@{postgresql.info.host}:"
                f"{postgresql.info.port}/{postgresql.info.dbname}"
            ),
            cache_files_urlpath=f"s3://{test_bucket_name}",
            cache_files_storage_options={"client_kwargs": client_kwargs},
        ):
            yield "cads"
    elif param.lower() in ("file", "local"):
        with config.set(
            cache_db_urlpath="sqlite:///" + str(tmp_path / "cacholote.db"),
            cache_files_urlpath=str(tmp_path / "cache_files"),
        ):
            yield "file"
    elif param.lower() == "off":
        yield "off"
    else:
        raise ValueError(f"{param=}")



"""Database objects."""

# Copyright 2023, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import datetime
import functools
import json
import os
import warnings
from typing import Any

import sqlalchemy as sa
import sqlalchemy.orm
import sqlalchemy_utils

import alembic.command
import alembic.config

from . import utils

_DATETIME_MAX = datetime.datetime(
    datetime.MAXYEAR, 12, 31, tzinfo=datetime.timezone.utc
)

Base = sa.orm.declarative_base()


class CacheEntry(Base):
    __tablename__ = "cache_entries"

    id = sa.Column(sa.Integer(), primary_key=True)
    key = sa.Column(sa.String(32))
    expiration = sa.Column(sa.DateTime, default=_DATETIME_MAX)
    result = sa.Column(sa.JSON)
    created_at = sa.Column(sa.DateTime, default=utils.utcnow)
    updated_at = sa.Column(sa.DateTime, default=utils.utcnow, onupdate=utils.utcnow)
    counter = sa.Column(sa.Integer)
    tag = sa.Column(sa.String)

    @property
    def _result_as_string(self) -> str:
        return json.dumps(self.result)

    def __repr__(self) -> str:
        public_attrs = (
            "id",
            "key",
            "expiration",
            "created_at",
            "updated_at",
            "counter",
            "tag",
        )
        public_attrs_repr = ", ".join(
            [f"{attr}={getattr(self, attr)!r}" for attr in public_attrs]
        )
        return f"CacheEntry({public_attrs_repr})"


@sa.event.listens_for(CacheEntry, "before_insert")
def set_expiration_to_max(
    mapper: sa.orm.Mapper[CacheEntry],
    connection: sa.Connection,
    target: CacheEntry,
) -> None:
    target.expiration = target.expiration or _DATETIME_MAX
    if target.expiration < utils.utcnow():
        warnings.warn(f"Expiration date has passed. {target.expiration=}", UserWarning)


def _commit_or_rollback(session: sa.orm.Session) -> None:
    try:
        session.commit()
    finally:
        session.rollback()


def _encode_kwargs(**kwargs: Any) -> dict[str, Any]:
    encoded_kwargs = {}
    for key, value in kwargs.items():
        if isinstance(value, dict):
            encoded_kwargs["_encoded_" + key] = json.dumps(value)
        else:
            encoded_kwargs[key] = value
    return encoded_kwargs


def _decode_kwargs(**kwargs: Any) -> dict[str, Any]:
    decoded_kwargs = {}
    for key, value in kwargs.items():
        if key.startswith("_encoded_"):
            decoded_kwargs[key.replace("_encoded_", "", 1)] = json.loads(value)
        else:
            decoded_kwargs[key] = value
    return decoded_kwargs


@functools.lru_cache
def _cached_sessionmaker(
    url: str, **kwargs: Any
) -> sa.orm.sessionmaker[sa.orm.Session]:
    engine = sa.create_engine(url, **_decode_kwargs(**kwargs))
    Base.metadata.create_all(engine)
    return sa.orm.sessionmaker(engine)


def cached_sessionmaker(url: str, **kwargs: Any) -> sa.orm.sessionmaker[sa.orm.Session]:
    return _cached_sessionmaker(url, **_encode_kwargs(**kwargs))


def init_database(connection_string: str, force: bool = False) -> sa.engine.Engine:
    """
    Make sure the db located at URI `connection_string` exists updated and return the engine object.

    :param connection_string: something like 'postgresql://user:password@netloc:port/dbname'
    :param force: if True, drop the database structure and build again from scratch
    """
    engine = sa.create_engine(connection_string)
    migration_directory = os.path.abspath(os.path.join(__file__, "..", ".."))
    os.chdir(migration_directory)
    alembic_config_path = os.path.join(migration_directory, "alembic.ini")
    alembic_cfg = alembic.config.Config(alembic_config_path)
    for option in ["drivername", "username", "password", "host", "port", "database"]:
        value = getattr(engine.url, option)
        if value is None:
            value = ""
        alembic_cfg.set_main_option(option, str(value))
    if not sqlalchemy_utils.database_exists(engine.url):
        sqlalchemy_utils.create_database(engine.url)
        # cleanup and create the schema
        Base.metadata.drop_all(engine)
        Base.metadata.create_all(engine)
        alembic.command.stamp(alembic_cfg, "head")
    else:
        # check the structure is empty or incomplete
        query = sa.text(
            "SELECT table_name FROM information_schema.tables WHERE table_schema='public'"
        )
        conn = engine.connect()
        if "cache_entries" not in conn.execute(query).scalars().all():
            force = True
        conn.close()
    if force:
        # cleanup and create the schema
        Base.metadata.drop_all(engine)
        Base.metadata.create_all(engine)
        alembic.command.stamp(alembic_cfg, "head")
    else:
        # update db structure
        alembic.command.upgrade(alembic_cfg, "head")
    return engine



"""Public decorator."""

# Copyright 2019, B-Open Solutions srl.
# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import functools
import json
import warnings
from typing import Any, Callable, TypeVar, cast

import sqlalchemy as sa
import sqlalchemy.orm

from . import clean, config, database, decode, encode, utils

F = TypeVar("F", bound=Callable[..., Any])


def _decode_and_update(
    session: sa.orm.Session,
    cache_entry: Any,
    settings: config.Settings,
) -> Any:
    result = decode.loads(cache_entry._result_as_string)
    cache_entry.counter = (cache_entry.counter or 0) + 1
    if settings.tag is not None:
        cache_entry.tag = settings.tag
    database._commit_or_rollback(session)
    if settings.return_cache_entry:
        session.refresh(cache_entry)
        return cache_entry
    return result


def cacheable(func: F) -> F:
    """Make a function cacheable."""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        settings = config.get()

        if not settings.use_cache and not settings.return_cache_entry:
            return func(*args, **kwargs)

        try:
            hexdigest = encode._hexdigestify_python_call(func, *args, **kwargs)
        except encode.EncodeError as ex:
            if settings.return_cache_entry:
                raise ex
            warnings.warn(f"can NOT encode python call: {ex!r}", UserWarning)
            return func(*args, **kwargs)

        if settings.use_cache:
            filters = [
                database.CacheEntry.key == hexdigest,
                database.CacheEntry.expiration > utils.utcnow(),
            ]
            if settings.expiration:
                # When expiration is provided, only get entries with matching expiration
                filters.append(database.CacheEntry.expiration == settings.expiration)

            with settings.instantiated_sessionmaker() as session:
                for cache_entry in session.scalars(
                    sa.select(database.CacheEntry)
                    .filter(*filters)
                    .order_by(database.CacheEntry.updated_at.desc())
                ):
                    try:
                        return _decode_and_update(session, cache_entry, settings)
                    except decode.DecodeError as ex:
                        warnings.warn(str(ex), UserWarning)
                        clean._delete_cache_entry(session, cache_entry)

        result = func(*args, **kwargs)
        cache_entry = database.CacheEntry(
            key=hexdigest,
            expiration=settings.expiration,
            tag=settings.tag,
        )
        try:
            cache_entry.result = json.loads(encode.dumps(result))
        except encode.EncodeError as ex:
            if settings.return_cache_entry:
                raise ex
            warnings.warn(f"can NOT encode output: {ex!r}", UserWarning)
            return result

        with settings.instantiated_sessionmaker() as session:
            session.add(cache_entry)
            return _decode_and_update(session, cache_entry, settings)

    return cast(F, wrapper)



"""add created_at and rename timestamp column.

Revision ID: a38663d192e5
Revises:
Create Date: 2024-07-24 15:53:43.989464

"""

from typing import Sequence, Union

import sqlalchemy as sa

import cacholote
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "a38663d192e5"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column(
        "cache_entries",
        sa.Column("created_at", sa.DateTime, default=cacholote.utils.utcnow),
    )
    op.execute(f"UPDATE cache_entries SET created_at='{cacholote.utils.utcnow()!s}'")

    op.alter_column("cache_entries", "timestamp", new_column_name="updated_at")


def downgrade() -> None:
    op.drop_column("cache_entries", "created_at")
    op.alter_column("cache_entries", "updated_at", new_column_name="timestamp")



"""Functions to clean cache database and files."""

# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import collections
import datetime
import posixpath
from typing import Any, Callable, Literal, Optional

import pydantic
import sqlalchemy as sa
import sqlalchemy.orm
from sqlalchemy import BinaryExpression, ColumnElement

from . import config, database, decode, encode, extra_encoders, utils

FILE_RESULT_KEYS = ("type", "callable", "args", "kwargs")
FILE_RESULT_CALLABLES = (
    "cacholote.extra_encoders:decode_xr_dataarray",
    "cacholote.extra_encoders:decode_xr_dataset",
    "cacholote.extra_encoders:decode_io_object",
)


def _get_files_from_cache_entry(cache_entry: database.CacheEntry) -> dict[str, str]:
    result = cache_entry.result
    if not isinstance(result, (list, tuple, set)):
        result = [result]

    files = {}
    for obj in result:
        if (
            isinstance(obj, dict)
            and set(FILE_RESULT_KEYS) == set(obj)
            and obj["callable"] in FILE_RESULT_CALLABLES
        ):
            fs, urlpath = extra_encoders._get_fs_and_urlpath(*obj["args"][:2])
            files[fs.unstrip_protocol(urlpath)] = obj["args"][0]["type"]
    return files


def _delete_cache_entry(
    session: sa.orm.Session, cache_entry: database.CacheEntry
) -> None:
    fs, _ = utils.get_cache_files_fs_dirname()
    files_to_delete = _get_files_from_cache_entry(cache_entry)
    logger = config.get().logger

    # First, delete database entry
    logger.info("deleting cache entry", cache_entry=cache_entry)
    session.delete(cache_entry)
    database._commit_or_rollback(session)

    # Then, delete files
    for urlpath, file_type in files_to_delete.items():
        if fs.exists(urlpath):
            logger.info("deleting cache file", urlpath=urlpath)
            fs.rm(urlpath, recursive=file_type == "application/vnd+zarr")


def delete(func_to_del: str | Callable[..., Any], *args: Any, **kwargs: Any) -> None:
    """Delete function previously cached.

    Parameters
    ----------
    func_to_del: callable, str
        Function to delete from cache
    *args: Any
        Argument of functions to delete from cache
    **kwargs: Any
        Keyword arguments of functions to delete from cache
    """
    hexdigest = encode._hexdigestify_python_call(func_to_del, *args, **kwargs)
    with config.get().instantiated_sessionmaker() as session:
        for cache_entry in session.scalars(
            sa.select(database.CacheEntry).filter(database.CacheEntry.key == hexdigest)
        ):
            _delete_cache_entry(session, cache_entry)


class _Cleaner:
    def __init__(self) -> None:
        self.logger = config.get().logger
        self.fs, self.dirname = utils.get_cache_files_fs_dirname()

        urldir = self.fs.unstrip_protocol(self.dirname)

        self.logger.info("getting disk usage")
        self.file_sizes: dict[str, int] = collections.defaultdict(int)
        for path, size in self.fs.du(self.dirname, total=False).items():
            # Group dirs
            urlpath = self.fs.unstrip_protocol(path)
            basename, *_ = urlpath.replace(urldir, "", 1).strip("/").split("/")
            if basename:
                self.file_sizes[posixpath.join(urldir, basename)] += size

        self.disk_usage = sum(self.file_sizes.values())
        self.log_disk_usage()

    def pop_file_size(self, file: str) -> int:
        size = self.file_sizes.pop(file, 0)
        self.disk_usage -= size
        return size

    def log_disk_usage(self) -> None:
        self.logger.info("check disk usage", disk_usage=self.disk_usage)

    def stop_cleaning(self, maxsize: int) -> bool:
        return self.disk_usage <= maxsize

    def get_unknown_files(self, lock_validity_period: float | None) -> set[str]:
        self.logger.info("getting unknown files")

        utcnow = utils.utcnow()
        locked_files = set()
        for urlpath in self.file_sizes:
            if urlpath.endswith(".lock"):
                modified = self.fs.modified(urlpath)
                if modified.tzinfo is None:
                    modified = modified.replace(tzinfo=datetime.timezone.utc)
                delta = utcnow - modified
                if lock_validity_period is None or delta < datetime.timedelta(
                    seconds=lock_validity_period
                ):
                    locked_files.add(urlpath)
                    locked_files.add(urlpath.rsplit(".lock", 1)[0])

        if unknown_files := (set(self.file_sizes) - locked_files):
            with config.get().instantiated_sessionmaker() as session:
                for cache_entry in session.scalars(sa.select(database.CacheEntry)):
                    for known_file in _get_files_from_cache_entry(cache_entry):
                        unknown_files.discard(known_file)
                    if not unknown_files:
                        break
        return unknown_files

    def delete_unknown_files(
        self, lock_validity_period: float | None, recursive: bool
    ) -> None:
        unknown_files = self.get_unknown_files(lock_validity_period)
        for urlpath in unknown_files:
            self.pop_file_size(urlpath)
        self.remove_files(
            list(unknown_files),
            recursive=recursive,
        )
        self.log_disk_usage()

    @staticmethod
    @pydantic.validate_call
    def _get_tag_filters(
        tags_to_clean: Optional[list[Optional[str]]],
        tags_to_keep: Optional[list[Optional[str]]],
    ) -> list[sa.ColumnElement[bool]]:
        if (tags_to_clean is not None) and (tags_to_keep is not None):
            raise ValueError("tags_to_clean/keep are mutually exclusive.")

        filters = []
        if tags_to_keep is not None:
            filters.append(
                sa.or_(
                    database.CacheEntry.tag.not_in(tags_to_keep),
                    database.CacheEntry.tag.is_not(None)
                    if None in tags_to_keep
                    else database.CacheEntry.tag.is_(None),
                )
            )
        elif tags_to_clean is not None:
            filters.append(
                sa.or_(
                    database.CacheEntry.tag.in_(tags_to_clean),
                    database.CacheEntry.tag.is_(None)
                    if None in tags_to_clean
                    else database.CacheEntry.tag.is_not(None),
                )
            )
        return filters

    @staticmethod
    @pydantic.validate_call
    def _get_method_sorters(
        method: Literal["LRU", "LFU"],
    ) -> list[sa.orm.InstrumentedAttribute[Any]]:
        sorters: list[sa.orm.InstrumentedAttribute[Any]] = []
        if method == "LRU":
            sorters.extend(
                [database.CacheEntry.updated_at, database.CacheEntry.counter]
            )
        elif method == "LFU":
            sorters.extend(
                [database.CacheEntry.counter, database.CacheEntry.updated_at]
            )
        else:
            raise ValueError(f"{method=}")
        sorters.append(database.CacheEntry.expiration)
        return sorters

    def remove_files(
        self,
        files: list[str],
        max_tries: int = 10,
        **kwargs: Any,
    ) -> None:
        assert max_tries >= 1
        if not files:
            return

        self.logger.info("deleting files", n_files_to_delete=len(files), **kwargs)

        n_tries = 0
        while files:
            n_tries += 1
            try:
                self.fs.rm(files, **kwargs)
                return
            except FileNotFoundError:
                # Another concurrent process might have deleted files
                if n_tries >= max_tries:
                    raise
                files = [file for file in files if self.fs.exists(file)]

    def delete_cache_files(
        self,
        maxsize: int,
        method: Literal["LRU", "LFU"],
        tags_to_clean: list[str | None] | None,
        tags_to_keep: list[str | None] | None,
    ) -> None:
        filters = self._get_tag_filters(tags_to_clean, tags_to_keep)
        sorters = self._get_method_sorters(method)

        if self.stop_cleaning(maxsize):
            return

        files_to_delete = []
        dirs_to_delete = []
        self.logger.info("getting cache entries to delete")
        n_entries_to_delete = 0
        with config.get().instantiated_sessionmaker() as session:
            for cache_entry in session.scalars(
                sa.select(database.CacheEntry).filter(*filters).order_by(*sorters)
            ):
                files = _get_files_from_cache_entry(cache_entry)
                if files:
                    n_entries_to_delete += 1
                    session.delete(cache_entry)

                for file, file_type in files.items():
                    self.pop_file_size(file)
                    if file_type == "application/vnd+zarr":
                        dirs_to_delete.append(file)
                    else:
                        files_to_delete.append(file)

                if self.stop_cleaning(maxsize):
                    break

            if n_entries_to_delete:
                self.logger.info(
                    "deleting cache entries", n_entries_to_delete=n_entries_to_delete
                )
            database._commit_or_rollback(session)

        self.remove_files(files_to_delete, recursive=False)
        self.remove_files(dirs_to_delete, recursive=True)
        self.log_disk_usage()

        if not self.stop_cleaning(maxsize):
            raise ValueError(
                (
                    f"Unable to clean {self.dirname!r}."
                    f" Final disk usage: {self.disk_usage!r}."
                    f" Target disk usage: {maxsize!r}"
                )
            )


def clean_cache_files(
    maxsize: int,
    method: Literal["LRU", "LFU"] = "LRU",
    delete_unknown_files: bool = False,
    recursive: bool = False,
    lock_validity_period: float | None = None,
    tags_to_clean: list[str | None] | None = None,
    tags_to_keep: list[str | None] | None = None,
) -> None:
    """Clean cache files.

    Parameters
    ----------
    maxsize: int
        Maximum total size of cache files (bytes).
    method: str, default: "LRU"
        * LRU: Last Recently Used
        * LFU: Least Frequently Used
    delete_unknown_files: bool, default: False
        Delete all files that are not registered in the cache database.
    recursive: bool, default: False
        Whether to delete unknown directories or not
    lock_validity_period: float, optional, default: None
        Validity period of lock files in seconds. Expired locks will be deleted.
    tags_to_clean, tags_to_keep: list of strings/None, optional, default: None
        Tags to clean/keep. If None, delete all cache entries.
        To delete/keep untagged entries, add None in the list (e.g., [None, 'tag1', ...]).
        tags_to_clean and tags_to_keep are mutually exclusive.
    """
    cleaner = _Cleaner()

    if delete_unknown_files:
        cleaner.delete_unknown_files(lock_validity_period, recursive)

    cleaner.delete_cache_files(
        maxsize=maxsize,
        method=method,
        tags_to_clean=tags_to_clean,
        tags_to_keep=tags_to_keep,
    )


def clean_invalid_cache_entries(
    check_expiration: bool = True, try_decode: bool = False
) -> None:
    """Clean invalid cache entries.

    Parameters
    ----------
    check_expiration: bool
        Whether or not to delete expired entries
    try_decode: bool
        Whether or not to delete entries that raise DecodeError (this can be slow!)
    """
    filters = []
    if check_expiration:
        filters.append(database.CacheEntry.expiration <= utils.utcnow())
    if filters:
        with config.get().instantiated_sessionmaker() as session:
            for cache_entry in session.scalars(
                sa.select(database.CacheEntry).filter(*filters)
            ):
                _delete_cache_entry(session, cache_entry)

    if try_decode:
        with config.get().instantiated_sessionmaker() as session:
            for cache_entry in session.scalars(sa.select(database.CacheEntry)):
                try:
                    decode.loads(cache_entry._result_as_string)
                except decode.DecodeError:
                    _delete_cache_entry(session, cache_entry)


def expire_cache_entries(
    tags: list[str] | None = None,
    before: datetime.datetime | None = None,
    after: datetime.date | None = None,
) -> int:
    now = utils.utcnow()

    filters: list[BinaryExpression[bool] | ColumnElement[bool]] = []
    if tags is not None:
        filters.append(database.CacheEntry.tag.in_(tags))
    if before is not None:
        filters.append(database.CacheEntry.created_at < before)
    if after is not None:
        filters.append(database.CacheEntry.created_at > after)

    count = 0
    with config.get().instantiated_sessionmaker() as session:
        for cache_entry in session.scalars(
            sa.select(database.CacheEntry).filter(*filters)
        ):
            count += 1
            cache_entry.expiration = now
        database._commit_or_rollback(session)
    return count



from __future__ import annotations

import datetime
from typing import Any

import pytest

from cacholote import config, decode, encode


def func(
    a: Any, b: Any, *args: Any, c: Any = None, d: Any = False, **kwargs: Any
) -> None:
    pass


def test_inspect_fully_qualified_name() -> None:
    # Goedel-style self-reference :)
    res = encode.inspect_fully_qualified_name(test_inspect_fully_qualified_name)
    assert res == "test_20_encode:test_inspect_fully_qualified_name"

    res = encode.inspect_fully_qualified_name(len)
    assert res == "builtins:len"


def test_dictify_python_object() -> None:
    res = encode.dictify_python_object(len)
    assert res == {"type": "python_object", "fully_qualified_name": "builtins:len"}

    expected = {
        "type": "python_object",
        "fully_qualified_name": "datetime:datetime.isoformat",
    }
    res = encode.dictify_python_object("datetime:datetime.isoformat")
    assert res == expected

    with pytest.raises(ValueError):
        encode.dictify_python_object("datetime.datetime.isoformat")


def test_dictify_python_call() -> None:
    expected0 = {"type": "python_call", "callable": "builtins:int"}
    res0 = encode.dictify_python_call(int)
    assert res0 == expected0

    expected1 = {"type": "python_call", "callable": "builtins:len", "args": ("test",)}
    res1 = encode.dictify_python_call(len, "test")
    assert res1 == expected1

    expected2 = {
        "type": "python_call",
        "callable": "builtins:int",
        "args": ("2",),
        "kwargs": {"base": 3},
    }
    res2 = encode.dictify_python_call(int, "2", base=3)
    assert res2 == expected2


def test_filecache_default() -> None:
    date = datetime.datetime.now()
    expected0 = {
        "type": "python_call",
        "callable": "datetime:datetime.fromisoformat",
        "args": (date.isoformat(),),
    }
    res0 = encode.filecache_default(date)

    assert res0 == expected0

    data1 = bytes(list(range(20)) + list(range(225, 256)))
    expected1 = {
        "type": "python_call",
        "callable": "binascii:a2b_base64",
        "args": (
            "AAECAwQFBgcICQoLDA0ODxAREhPh4uPk5ebn6Onq6+zt7u/w8fLz9PX29/j5+vv8/f7/\n",
        ),
    }
    res1 = encode.filecache_default(data1)
    assert res1 == expected1

    data2 = datetime.timezone(datetime.timedelta(seconds=3600))
    expected2 = {
        "type": "python_call",
        "callable": "_pickle:loads",
        "args": (
            b"\x80\x04\x958\x00\x00\x00\x00\x00\x00\x00\x8c\x08"
            b"datetime\x94\x8c\x08timezone\x94\x93\x94h\x00\x8c\t"
            b"timedelta\x94\x93\x94K\x00M\x10\x0eK\x00\x87\x94R\x94\x85\x94R\x94.",
        ),
    }
    res2 = encode.filecache_default(data2)
    assert res2 == expected2


@pytest.mark.parametrize("raise_all_encoding_errors", [True, False])
def test_filecache_default_error(raise_all_encoding_errors: bool) -> None:
    config.set(raise_all_encoding_errors=raise_all_encoding_errors)

    class Dummy:
        pass

    if raise_all_encoding_errors:
        with pytest.raises(AttributeError, match="Can't pickle local object"):
            encode.filecache_default(Dummy())
    else:
        with pytest.warns(UserWarning, match="Can't pickle local object"):
            with pytest.raises(encode.EncodeError):
                encode.filecache_default(Dummy())


@pytest.mark.parametrize(
    "data",
    [
        len,
        bytes(list(range(255))),
        datetime.datetime.now(),
        datetime.date.today(),
        datetime.timedelta(234, 23, 128736),
        datetime.timezone(datetime.timedelta(seconds=3600)),  # pickle encode / decode
    ],
    ids=[f"data{i}" for i in range(6)],
)
def test_roundtrip(data: Any) -> None:
    assert decode.loads(encode.dumps(data)) == data


def test_dumps_python_call() -> None:
    expected = (
        r'{"type":"python_call","callable":"datetime:datetime",'
        r'"args":[2019,1,1],"kwargs":{"tzinfo":{"type":"python_call",'
        r'"callable":"_pickle:loads","args":[{"type":"python_call",'
        r'"callable":"binascii:a2b_base64","args":["gASVOAAAAAAAAACMCGRhdGV0aW1ll'
        r'IwIdGltZXpvbmWUk5RoAIwJdGltZWRlbHRhlJOUSwBNEA5LAIeUUpSFlFKULg==\n"]}]}}}'
    )
    tzinfo = datetime.timezone(datetime.timedelta(seconds=3600))
    res = encode.dumps_python_call("datetime:datetime", 2019, 1, 1, tzinfo=tzinfo)
    assert res == expected

    res_decoded = decode.loads(res)
    assert res_decoded == datetime.datetime(2019, 1, 1, tzinfo=tzinfo)


def test_dumps_json_serializable() -> None:
    expected = "1"
    actual = encode.dumps(1)
    assert expected == actual



"""Global settings."""

# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import abc
import datetime
import logging
import pathlib
import tempfile
from types import TracebackType
from typing import Any, Literal, Optional, Union

import fsspec
import pydantic
import pydantic_settings
import sqlalchemy as sa
import sqlalchemy.orm
import structlog

from . import database

_SETTINGS: Settings | None = None
_DEFAULT_CACHE_DIR = pathlib.Path(tempfile.gettempdir()) / "cacholote"
_DEFAULT_CACHE_DIR.mkdir(exist_ok=True)
_DEFAULT_CACHE_DB_URLPATH = f"sqlite:///{_DEFAULT_CACHE_DIR / 'cacholote.db'}"
_DEFAULT_CACHE_FILES_URLPATH = f"{_DEFAULT_CACHE_DIR / 'cache_files'}"
_DEFAULT_LOGGER = structlog.get_logger(
    wrapper_class=structlog.make_filtering_bound_logger(logging.WARNING)
)


class Context(abc.ABC):
    @abc.abstractmethod
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

    @abc.abstractmethod
    def upload_log(self, *args: Any, **kwargs: Any) -> None: ...


class Settings(pydantic_settings.BaseSettings):
    use_cache: bool = True
    cache_db_urlpath: Optional[str] = _DEFAULT_CACHE_DB_URLPATH
    create_engine_kwargs: dict[str, Any] = {}
    sessionmaker: Optional[sa.orm.sessionmaker[sa.orm.Session]] = None
    cache_files_urlpath: str = _DEFAULT_CACHE_FILES_URLPATH
    cache_files_urlpath_readonly: Optional[str] = None
    cache_files_storage_options: dict[str, Any] = {}
    xarray_cache_type: Literal[
        "application/netcdf", "application/x-grib", "application/vnd+zarr"
    ] = "application/netcdf"
    io_delete_original: bool = False
    raise_all_encoding_errors: bool = False
    expiration: Optional[datetime.datetime] = None
    tag: Optional[str] = None
    return_cache_entry: bool = False
    logger: Union[structlog.BoundLogger, structlog._config.BoundLoggerLazyProxy] = (
        _DEFAULT_LOGGER
    )
    lock_timeout: Optional[float] = None
    context: Optional[Context] = None

    @pydantic.field_validator("create_engine_kwargs")
    def validate_create_engine_kwargs(
        cls: pydantic_settings.BaseSettings, create_engine_kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        poolclass = create_engine_kwargs.get("poolclass")
        if isinstance(poolclass, str):
            create_engine_kwargs["poolclass"] = getattr(sa.pool, poolclass)
        return create_engine_kwargs

    @pydantic.field_validator("expiration")
    def validate_expiration(
        cls: pydantic_settings.BaseSettings, expiration: datetime.datetime | None
    ) -> datetime.datetime | None:
        if expiration is not None and expiration.tzinfo is None:
            raise ValueError(f"Expiration is missing the timezone info. {expiration=}")
        return expiration

    @pydantic.model_validator(mode="after")
    def make_cache_dir(self) -> Settings:
        fs, _, (urlpath, *_) = fsspec.get_fs_token_paths(
            self.cache_files_urlpath,
            storage_options=self.cache_files_storage_options,
        )
        fs.mkdirs(urlpath, exist_ok=True)
        return self

    @property
    def instantiated_sessionmaker(self) -> sa.orm.sessionmaker[sa.orm.Session]:
        if self.sessionmaker is None:
            if self.cache_db_urlpath is None:
                raise ValueError("Provide either `sessionmaker` or `cache_db_urlpath`.")
            self.sessionmaker = database.cached_sessionmaker(
                self.cache_db_urlpath, **self.create_engine_kwargs
            )
            self.cache_db_urlpath = None
            self.create_engine_kwargs = {}
        elif self.cache_db_urlpath is not None:
            raise ValueError(
                "`sessionmaker` and `cache_db_urlpath` are mutually exclusive."
            )
        return self.sessionmaker

    @property
    def engine(self) -> sa.engine.Engine:
        engine = self.instantiated_sessionmaker.kw["bind"]
        assert isinstance(engine, sa.engine.Engine)
        return engine

    model_config = pydantic_settings.SettingsConfigDict(
        case_sensitive=False, env_prefix="cacholote_"
    )


class set:
    """Customize cacholote settings.

    It is possible to use it either as a context manager, or to configure global settings.

    Parameters
    ----------
    use_cache: bool, default: True
        Enable/disable cache.
    cache_db_urlpath: str, None, default:"sqlite:////system_tmp_dir/cacholote/cacholote.db"
        URL for cache database (driver://user:pass@host/database).
    create_engine_kwargs: dict, default: {}
        Keyword arguments for ``sqlalchemy.create_engine``
    sessionmaker: sessionmaker, optional
        sqlalchemy.sessionamaker, mutually exclusive with cache_db_urlpath and create_engine_kwargs
    cache_files_urlpath: str, default:"/system_tmp_dir/cacholote/cache_files"
        URL for cache files (protocol://location).
    cache_files_storage_options: dict, default: {}
        ``fsspec`` storage options for storing cache files.
    cache_files_urlpath_readonly: str, None, default: None
        URL for cache files accessible in read-only mode.
        None: same as ``cache_files_urlpath``
    xarray_cache_type: {"application/netcdf", "application/x-grib", "application/vnd+zarr"}, \
        default: "application/netcdf"
        Type for ``xarray`` cache files.
    io_delete_original: bool, default: False
        Whether to delete the original copy of cached files.
    raise_all_encoding_errors: bool, default: False
        Raise an error if an encoder does not work (i.e., do not return results).
    expiration: datetime, optional, default: None
        Expiration for cached results.
    tag: str, optional, default: None
        Tag for the cache entry. If None, do NOT tag.
        Note that existing tags are overwritten.
    return_cache_entry: bool, default: False
        Whether to return the cache database entry rather than decoded results.
    lock_timeout: float, optional, default: None
        Time to wait before raising an error if a cache file is locked.
    context: Context, optional, default: None
        CADS context for internal use.
    """

    def __init__(self, **kwargs: Any):
        self._old_settings = get()

        model_dump = self._old_settings.model_dump()
        if kwargs.get("cache_db_urlpath"):
            model_dump["sessionmaker"] = None
        if kwargs.get("sessionmaker"):
            model_dump["cache_db_urlpath"] = None
        model_dump.update(kwargs)

        global _SETTINGS
        _SETTINGS = Settings(**model_dump)

    def __enter__(self) -> Settings:
        return get()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        global _SETTINGS
        _SETTINGS = self._old_settings


def reset(env_file: str | tuple[str] | None = None) -> None:
    """Reset cacholote settings.

    Priority:
    1. Environment variables with prefix `CACHOLOTE_`
    2. Dotenv file(s)
    3. Cacholote defaults

    Parameters
    ----------
    env_file: str, tuple[str], default=None
        Dot env file(s).
    """
    global _SETTINGS
    _SETTINGS = Settings(_env_file=env_file)  # type: ignore[call-arg]
    set()


def get() -> Settings:
    """Get cacholote settings."""
    if _SETTINGS is None:
        reset()
        assert _SETTINGS is not None, "reset() did not work properly"
    return _SETTINGS.model_copy()



"""Additional encoders that need optional dependencies."""

# Copyright 2019, B-Open Solutions srl.
# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import contextlib
import functools
import hashlib
import inspect
import io
import mimetypes
import pathlib
import posixpath
import tempfile
import time
from collections.abc import Generator
from typing import (
    Any,
    Callable,
    Literal,
    TypeVar,
    Union,
    cast,
    overload,
)

import fsspec
import fsspec.implementations.local
import pydantic

from . import config, encode, utils

try:
    import dask
    import xarray as xr

    _HAS_XARRAY_AND_DASK = True
except ImportError:
    _HAS_XARRAY_AND_DASK = False

try:
    import magic

    _HAS_MAGIC = True
except ImportError:
    _HAS_MAGIC = False

F = TypeVar("F", bound=Callable[..., Any])

_UNION_IO_TYPES = Union[
    io.RawIOBase,
    io.BufferedIOBase,
    io.TextIOBase,
    fsspec.spec.AbstractBufferedFile,
    fsspec.implementations.local.LocalFileOpener,
]


def _add_ext_to_mimetypes() -> None:
    """Add netcdf, grib, and zarr to mimetypes."""
    mimetypes.add_type("application/netcdf", ".nc", strict=True)
    for ext in (".grib", ".grb", ".grb1", ".grb2"):
        mimetypes.add_type("application/x-grib", ext, strict=False)
    mimetypes.add_type("application/vnd+zarr", ".zarr", strict=False)


_add_ext_to_mimetypes()


def _guess_type(
    fs: fsspec.AbstractFileSystem,
    local_path: str,
    default: str = "application/octet-stream",
) -> str:
    if content_type := fs.info(local_path).get("ContentType", ""):
        return str(content_type)

    filetype, *_ = mimetypes.guess_type(local_path, strict=False)
    if filetype is None and _HAS_MAGIC:
        with fs.open(local_path, "rb") as f:
            filetype = magic.from_buffer(f.read(), mime=True)
    return filetype or default


def _requires_xarray_and_dask(func: F) -> F:
    """Raise an error if `xarray` or `dask` are not installed."""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if not _HAS_XARRAY_AND_DASK:
            raise ValueError("please install 'xarray' and 'dask'")
        return func(*args, **kwargs)

    return cast(F, wrapper)


def _filesystem_is_local(fs: fsspec.AbstractFileSystem) -> bool:
    return isinstance(fs, fsspec.get_filesystem_class("file"))


def _kwargs_to_str(**kwargs: Any) -> str:
    return " ".join([f"{k}={v}" for k, v in kwargs.items()])


@contextlib.contextmanager
def _logging_timer(event: str, **kwargs: Any) -> Generator[float, None, None]:
    logger = config.get().logger
    context = config.get().context
    logger.info(f"start {event}", **kwargs)
    if event == "upload" and context is not None:
        context.upload_log(f"start {event}. {_kwargs_to_str(**kwargs)}")

    tic = time.perf_counter()
    yield tic
    toc = time.perf_counter()

    kwargs["_".join(event.split() + ["time"])] = toc - tic  # elapsed time
    logger.info(f"end {event}", **kwargs)
    if event == "upload" and context is not None:
        context.upload_log(f"end {event}. {_kwargs_to_str(**kwargs)}")


class FileInfoModel(pydantic.BaseModel):
    type: str
    href: str
    file_checksum: str = pydantic.Field(..., alias="file:checksum")
    file_size: int = pydantic.Field(..., alias="file:size")
    file_local_path: str = pydantic.Field(..., alias="file:local_path")


def _dictify_file(fs: fsspec.AbstractFileSystem, local_path: str) -> dict[str, Any]:
    settings = config.get()
    href = posixpath.join(
        settings.cache_files_urlpath_readonly or settings.cache_files_urlpath,
        posixpath.basename(local_path),
    )
    file_dict = {
        "type": _guess_type(fs, local_path),
        "href": href,
        "file:checksum": f"{fs.checksum(local_path):x}",
        "file:size": fs.size(local_path),
        "file:local_path": local_path,
    }
    return FileInfoModel(**file_dict).model_dump(by_alias=True)


def _get_fs_and_urlpath(
    file_json: dict[str, Any],
    storage_options: dict[str, Any] | None = None,
    validate: bool = False,
) -> tuple[fsspec.AbstractFileSystem, str]:
    settings = config.get()
    urlpath = file_json["file:local_path"]
    if storage_options is None:
        storage_options = settings.cache_files_storage_options

    if not validate:
        fs, *_ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
        return (fs, urlpath)

    # Attempt to read from local_path
    try:
        fs, *_ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
    except:  # noqa: E722
        pass
    else:
        if fs.exists(urlpath):
            expected = file_json["file:checksum"]
            actual = (
                fs.checksum(urlpath)  # Just for backward compatibility.
                if isinstance(expected, int)
                else f"{fs.checksum(urlpath):x}"
            )
            if expected != actual:
                raise ValueError(f"checksum mismatch: {urlpath=} {expected=} {actual=}")
            settings.logger.info(
                "retrieve cache file", urlpath=fs.unstrip_protocol(urlpath)
            )
            return (fs, urlpath)

    # Attempt to read from href
    urlpath = file_json["href"]
    fs, *_ = fsspec.get_fs_token_paths(urlpath)
    if fs.exists(urlpath):
        return (fs, urlpath)

    # Nothing worked
    raise ValueError(
        f"No such file or directory: {file_json['file:local_path']!r} nor {file_json['href']!r}"
    )


@overload
def decode_xr_object(
    file_json: dict[str, Any],
    storage_options: dict[str, Any],
    xr_type: Literal["DataArray"],
    **kwargs: Any,
) -> xr.DataArray: ...


@overload
def decode_xr_object(
    file_json: dict[str, Any],
    storage_options: dict[str, Any],
    xr_type: Literal["Dataset"],
    **kwargs: Any,
) -> xr.Dataset: ...


@_requires_xarray_and_dask
def decode_xr_object(
    file_json: dict[str, Any],
    storage_options: dict[str, Any],
    xr_type: Literal["Dataset", "DataArray"],
    **kwargs: Any,
) -> xr.Dataset | xr.DataArray:
    fs, urlpath = _get_fs_and_urlpath(
        file_json, storage_options=storage_options, validate=True
    )

    if file_json["type"] == "application/vnd+zarr":
        filename_or_obj = fs.get_mapper(urlpath)
    else:
        if _filesystem_is_local(fs):
            filename_or_obj = urlpath
        else:
            # Download local copy
            protocols = (fs.protocol,) if isinstance(fs.protocol, str) else fs.protocol
            with fsspec.open(
                f"filecache::{urlpath}",
                filecache={"same_names": True},
                **{protocol: fs.storage_options for protocol in protocols},
            ) as of:
                filename_or_obj = of.name
    if xr_type == "Dataset":
        return xr.open_dataset(filename_or_obj, **kwargs)
    return xr.open_dataarray(filename_or_obj, **kwargs)


def decode_xr_dataset(
    file_json: dict[str, Any], storage_options: dict[str, Any], **kwargs: Any
) -> xr.Dataset:
    return decode_xr_object(file_json, storage_options, "Dataset", **kwargs)


def decode_xr_dataarray(
    file_json: dict[str, Any], storage_options: dict[str, Any], **kwargs: Any
) -> xr.DataArray:
    return decode_xr_object(file_json, storage_options, "DataArray", **kwargs)


def decode_io_object(
    file_json: dict[str, Any], storage_options: dict[str, Any], **kwargs: Any
) -> _UNION_IO_TYPES:
    fs, urlpath = _get_fs_and_urlpath(
        file_json, storage_options=storage_options, validate=True
    )
    return fs.open(urlpath, **kwargs)


@_requires_xarray_and_dask
def _store_xr_object(
    obj: xr.Dataset | xr.DataArray,
    fs: fsspec.AbstractFileSystem,
    urlpath: str,
    filetype: str,
) -> None:
    if filetype == "application/vnd+zarr":
        # Write directly on any filesystem
        mapper = fs.get_mapper(urlpath)
        with _logging_timer("upload", urlpath=fs.unstrip_protocol(urlpath)):
            obj.to_zarr(mapper, consolidated=True)
        return

    # Need a tmp local copy to write on a different filesystem
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpfilename = str(pathlib.Path(tmpdirname) / pathlib.Path(urlpath).name)

        with _logging_timer("write tmp file", urlpath=tmpfilename):
            if filetype == "application/netcdf":
                obj.to_netcdf(tmpfilename)
            elif filetype == "application/x-grib":
                import cfgrib.xarray_to_grib

                cfgrib.xarray_to_grib.to_grib(obj, tmpfilename)
            else:
                # Should never get here! xarray_cache_type is checked in config.py
                raise ValueError(f"type {filetype!r} is NOT supported.")

        _store_file_object(
            fs if _filesystem_is_local(fs) else fsspec.filesystem("file"),
            tmpfilename,
            fs,
            urlpath,
            io_delete_original=True,
        )


@_requires_xarray_and_dask
def dictify_xr_object(obj: xr.Dataset | xr.DataArray) -> dict[str, Any]:
    """Encode a ``xr.Dataset`` to JSON deserialized data (``dict``)."""
    settings = config.get()
    with dask.config.set({"tokenize.ensure-deterministic": True}):
        root = dask.base.tokenize(obj)

    ext = mimetypes.guess_extension(settings.xarray_cache_type, strict=False)
    urlpath_out = posixpath.join(settings.cache_files_urlpath, f"{root}{ext}")

    fs_out, *_ = fsspec.get_fs_token_paths(
        urlpath_out,
        storage_options=settings.cache_files_storage_options,
    )
    with utils.FileLock(
        fs_out, urlpath_out, timeout=settings.lock_timeout
    ) as file_exists:
        if not file_exists:
            _store_xr_object(obj, fs_out, urlpath_out, settings.xarray_cache_type)

        file_json = _dictify_file(fs_out, urlpath_out)

        kwargs: dict[str, Any] = {"chunks": {}}
        if settings.xarray_cache_type == "application/vnd+zarr":
            kwargs.update({"engine": "zarr", "consolidated": True})

        return encode.dictify_python_call(
            decode_xr_dataset if isinstance(obj, xr.Dataset) else decode_xr_dataarray,
            file_json,
            storage_options=settings.cache_files_storage_options,
            **kwargs,
        )


def _store_file_object(
    fs_in: fsspec.AbstractFileSystem,
    urlpath_in: str,
    fs_out: fsspec.AbstractFileSystem,
    urlpath_out: str,
    io_delete_original: bool | None = None,
) -> None:
    if io_delete_original is None:
        io_delete_original = config.get().io_delete_original

    kwargs = {}
    if content_type := _guess_type(fs_in, urlpath_in):
        kwargs["ContentType"] = content_type
    with _logging_timer(
        "upload",
        urlpath=fs_out.unstrip_protocol(urlpath_out),
        size=fs_in.size(urlpath_in),
    ):
        if fs_in == fs_out:
            func = fs_in.mv if io_delete_original else fs_in.cp
            func(urlpath_in, urlpath_out, **kwargs)
        elif _filesystem_is_local(fs_in):
            fs_out.put(urlpath_in, urlpath_out, **kwargs)
        else:
            with fs_in.open(urlpath_in, "rb") as f_in:
                with fs_out.open(urlpath_out, "wb") as f_out:
                    utils.copy_buffered_file(f_in, f_out)

    if io_delete_original and fs_in.exists(urlpath_in):
        with _logging_timer(
            "remove",
            urlpath=fs_in.unstrip_protocol(urlpath_in),
            size=fs_in.size(urlpath_in),
        ):
            fs_in.rm(urlpath_in)


def _store_io_object(
    f_in: _UNION_IO_TYPES,
    fs_out: fsspec.AbstractFileSystem,
    urlpath_out: str,
) -> None:
    f_out = fs_out.open(urlpath_out, "wb")
    with _logging_timer("upload", urlpath=fs_out.unstrip_protocol(urlpath_out)):
        utils.copy_buffered_file(f_in, f_out)


def dictify_io_object(obj: _UNION_IO_TYPES) -> dict[str, Any]:
    """Encode a file object to JSON deserialized data (``dict``)."""
    settings = config.get()

    cache_files_urlpath = settings.cache_files_urlpath
    fs_out, *_ = fsspec.get_fs_token_paths(
        cache_files_urlpath,
        storage_options=settings.cache_files_storage_options,
    )

    if urlpath_in := getattr(obj, "path", getattr(obj, "name", "")):
        fs_in = getattr(obj, "fs", fsspec.filesystem("file"))
        root = f"{fs_in.checksum(urlpath_in):x}"
        ext = pathlib.Path(urlpath_in).suffix
        urlpath_out = posixpath.join(cache_files_urlpath, f"{root}{ext}")
    else:
        root = hashlib.md5(f"{hash(obj)}".encode()).hexdigest()  # fsspec uses md5
        urlpath_out = posixpath.join(cache_files_urlpath, root)

    with utils.FileLock(
        fs_out, urlpath_out, timeout=settings.lock_timeout
    ) as file_exists:
        if not file_exists:
            if urlpath_in:
                _store_file_object(fs_in, urlpath_in, fs_out, urlpath_out)
            else:
                _store_io_object(obj, fs_out, urlpath_out)

        file_json = _dictify_file(fs_out, urlpath_out)

        params = inspect.signature(open).parameters
        kwargs = {k: getattr(obj, k) for k in params.keys() if hasattr(obj, k)}

        return encode.dictify_python_call(
            decode_io_object,
            file_json,
            storage_options=settings.cache_files_storage_options,
            **kwargs,
        )


def register_all() -> None:
    """Register extra encoders if optional dependencies are installed."""
    for type_ in (
        io.RawIOBase,
        io.BufferedIOBase,
        io.TextIOBase,
        fsspec.spec.AbstractBufferedFile,
        fsspec.implementations.local.LocalFileOpener,
    ):
        encode.FILECACHE_ENCODERS.append((type_, dictify_io_object))
    if _HAS_XARRAY_AND_DASK:
        for type_ in (xr.Dataset, xr.DataArray):
            encode.FILECACHE_ENCODERS.append((type_, dictify_xr_object))



"""module for entry points."""

import sqlalchemy as sa

import alembic.context
import cacholote

config = alembic.context.config


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url_props = dict()
    for prop in ["drivername", "username", "password", "host", "port", "database"]:
        url_props[prop] = config.get_main_option(prop)
    url_props["port"] = url_props["port"] and int(url_props["port"]) or None  # type: ignore
    url = sa.engine.URL.create(**url_props)  # type: ignore
    alembic.context.configure(
        url=url,
        target_metadata=cacholote.database.Base.metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )
    with alembic.context.begin_transaction():
        alembic.context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    url_props = dict()
    for prop in ["drivername", "username", "password", "host", "port", "database"]:
        url_props[prop] = config.get_main_option(prop)
    url_props["port"] = url_props["port"] and int(url_props["port"]) or None  # type: ignore
    url = sa.engine.URL.create(**url_props)  # type: ignore
    engine = sa.create_engine(url, poolclass=sa.pool.NullPool)
    with engine.connect() as connection:
        alembic.context.configure(
            connection=connection,
            target_metadata=cacholote.database.Base.metadata,
            version_table="alembic_version_cacholote",
        )

        with alembic.context.begin_transaction():
            alembic.context.run_migrations()


if alembic.context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""Utilities."""

# Copyright 2019, B-Open Solutions srl.
# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import hashlib
from __future__ import annotations

import dataclasses
import datetime
import functools
import hashlib
import io
import time
import warnings
from types import TracebackType
from typing import Any

import fsspec

from . import config


def hexdigestify(text: str) -> str:
    """Convert text to its hash made of hexadecimal digits."""
    hash_req = hashlib.sha3_224(text.encode())
    return hash_req.hexdigest()[:32]


def get_cache_files_fs_dirname() -> tuple[fsspec.AbstractFileSystem, str]:
    """Return the ``fsspec`` filesystem and directory name where cache files are stored."""
    fs, _, (path,) = fsspec.get_fs_token_paths(
        config.get().cache_files_urlpath,
        storage_options=config.get().cache_files_storage_options,
    )
    fs.invalidate_cache()
    return (fs, path)


def copy_buffered_file(
    f_in: Any,
    f_out: fsspec.spec.AbstractBufferedFile,
    buffer_size: int | None = None,
) -> None:
    """Copy file in chunks.

    Parameters
    ----------
    f_in, f_out: fsspec.spec.AbstractBufferedFile
        Source and destination buffered files.
    buffer_size: int, optional, default=None
        Maximum size for chunks in bytes.
        None: use ``io.DEFAULT_BUFFER_SIZE``
    """
    if buffer_size is None:
        buffer_size = io.DEFAULT_BUFFER_SIZE
    while True:
        data = f_in.read(buffer_size)
        if not data:
            break
        f_out.write(data if isinstance(data, bytes) else data.encode())


@dataclasses.dataclass
class FileLock:
    fs: fsspec.AbstractFileSystem  # fsspec file system
    urlpath: str  # file to lock
    timeout: float | None  # lock timeout in seconds

    @functools.cached_property
    def lockfile(self) -> str:
        return self.urlpath + ".lock"

    def acquire(self) -> None:
        self.fs.touch(self.lockfile)

    @property
    def exists(self) -> bool:
        return bool(self.fs.exists(self.urlpath))

    @property
    def lock_exists(self) -> bool:
        return bool(self.fs.exists(self.lockfile))

    def release(self) -> None:
        if self.fs.exists(self.lockfile):
            self.fs.rm(self.lockfile)

    @property
    def is_locked(self) -> bool:
        return self.lock_exists and self.exists

    def wait_until_released(self) -> None:
        warned = False
        message = f"{self.urlpath!r} is locked: {self.lockfile!r}"
        start = time.perf_counter()
        while self.is_locked:
            if self.timeout is not None and time.perf_counter() - start > self.timeout:
                raise TimeoutError(message)
            if not warned:
                warnings.warn(message, UserWarning)
                warned = True
            time.sleep(min(1, self.timeout or 1))

    def __enter__(self) -> bool:
        self.wait_until_released()
        self.acquire()
        return self.exists

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.release()


def utcnow() -> datetime.datetime:
    """See https://discuss.python.org/t/deprecating-utcnow-and-utcfromtimestamp/26221."""
    return datetime.datetime.now(tz=datetime.timezone.utc)



"""Encode to JSON data."""

# Copyright 2019, B-Open Solutions srl.
# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import binascii
import collections.abc
import datetime
import inspect
import json
import pickle
import warnings
from typing import Any, Callable

from . import config, decode, utils

_JSON_DUMPS_KWARGS: dict[str, Any] = {"separators": (",", ":"), "skipkeys": False}


def _hexdigestify_python_call(
    func_to_hex: str | Callable[..., Any],
    *args: Any,
    **kwargs: Any,
) -> str:
    return utils.hexdigestify(dumps_python_call(func_to_hex, *args, **kwargs))


def inspect_fully_qualified_name(obj: Callable[..., Any]) -> str:
    """Return the fully qualified name of a python object."""
    module = inspect.getmodule(obj)
    if module is None:
        raise ValueError(f"can't getmodule for {obj!r}")
    return f"{module.__name__}:{obj.__qualname__}"


def dictify_python_object(obj: str | Callable[..., Any]) -> dict[str, str]:
    if isinstance(obj, str):
        # NOTE: a stricter test would be decode.import_object(obj)
        if ":" not in obj:
            raise ValueError(f"{obj} not in the form 'module:qualname'")
        fully_qualified_name = obj
    else:
        fully_qualified_name = inspect_fully_qualified_name(obj)
    return {
        "type": "python_object",
        "fully_qualified_name": fully_qualified_name,
    }


def dictify_python_call(
    func_to_dict: str | Callable[..., Any],
    *args: Any,
    **kwargs: Any,
) -> dict[str, Any]:
    callable_fqn = dictify_python_object(func_to_dict)["fully_qualified_name"]
    python_call_simple: dict[str, Any] = {
        "type": "python_call",
        "callable": callable_fqn,
    }

    callable_obj = (
        decode.import_object(callable_fqn)
        if isinstance(func_to_dict, str)
        else func_to_dict
    )
    try:
        sig = inspect.signature(callable_obj)
    except ValueError:
        # No signature available
        pass
    else:
        bound = sig.bind(*args, **kwargs)
        args = bound.args
        kwargs = bound.kwargs
    if args:
        python_call_simple["args"] = args
    if kwargs:
        python_call_simple["kwargs"] = {k: kwargs[k] for k in sorted(kwargs)}

    return python_call_simple


def dictify_datetime(obj: datetime.datetime) -> dict[str, Any]:
    # Work around "AttributeError: 'NoneType' object has no attribute '__name__'"
    return dictify_python_call("datetime:datetime.fromisoformat", obj.isoformat())


def dictify_date(obj: datetime.date) -> dict[str, Any]:
    return dictify_python_call("datetime:date.fromisoformat", obj.isoformat())


def dictify_timedelta(obj: datetime.timedelta) -> dict[str, Any]:
    return dictify_python_call(
        "datetime:timedelta", obj.days, obj.seconds, obj.microseconds
    )


def dictify_bytes(obj: bytes) -> dict[str, Any]:
    ascii_decoded = binascii.b2a_base64(obj).decode("ascii")
    return dictify_python_call(binascii.a2b_base64, ascii_decoded)


def dictify_pickable(obj: Any) -> dict[str, Any]:
    return dictify_python_call(pickle.loads, pickle.dumps(obj))


FILECACHE_ENCODERS: list[tuple[Any, Callable[[Any], dict[str, Any]]]] = [
    (object, dictify_pickable),
    (collections.abc.Callable, dictify_python_object),
    (bytes, dictify_bytes),
    (datetime.date, dictify_date),
    (datetime.datetime, dictify_datetime),
    (datetime.timedelta, dictify_timedelta),
]


class EncodeError(Exception):
    pass


def filecache_default(
    obj: Any,
    encoders: list[tuple[Any, Callable[[Any], dict[str, Any]]]] | None = None,
) -> dict[str, Any]:
    """Dictify objects that are not JSON-serializable.

    Parameters
    ----------
    obj: Any
        Object to encode
    encoders: list, optional
        List of tuples of the form ``(type, encoder)``.
        None: Use default ``cacholote.encode.FILECACHE_ENCODERS``

    Returns
    -------
    dict
    """
    if encoders is None:
        encoders = FILECACHE_ENCODERS
    for type_, encoder in reversed(encoders):
        if isinstance(obj, type_):
            try:
                return encoder(obj)
            except Exception as ex:
                if config.get().raise_all_encoding_errors:
                    raise ex
                warnings.warn(f"{encoder!r} did not work: {ex!r}")
    raise EncodeError("can't encode object")


def dumps(
    obj: Any,
    **kwargs: Any,
) -> str:
    """Serialize object to JSON formatted string.

    Parameters
    ----------
    obj: Any
        Object to serialize
    **kwargs:
        Keyword arguments of ``json.dumps``

    Returns
    -------
    str
    """
    for key, value in _JSON_DUMPS_KWARGS.items():
        kwargs.setdefault(key, value)
    kwargs.setdefault("default", filecache_default)

    return json.dumps(obj, **kwargs)


def dumps_python_call(
    func_to_dump: str | Callable[..., Any],
    *args: Any,
    **kwargs: Any,
) -> str:
    """Serialize python call to JSON formatted string.

    Parameters
    ----------
    func_to_dump: str, callable
        Function to serialize
    *args: Any
        Arguments of ``func``
    **kwargs: Any
        Keyword arguments of ``func``

    Returns
    -------
    str
    """
    python_call = dictify_python_call(func_to_dump, *args, **kwargs)
    return dumps(python_call)



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import fsspec
import pytest
import pytest_structlog
import structlog

from cacholote import cache, config, decode, encode, extra_encoders, utils

if TYPE_CHECKING:
    import xarray as xr
else:
    xr = pytest.importorskip("xarray")
dask = pytest.importorskip("dask")


def get_grib_ds() -> xr.Dataset:
    pytest.importorskip("cfgrib")
    eccodes = pytest.importorskip("eccodes")
    filename = pathlib.Path(eccodes.codes_samples_path()) / "GRIB2.tmpl"
    ds = xr.open_dataset(filename, engine="cfgrib")
    del ds.attrs["history"]
    return ds


@pytest.mark.filterwarnings(
    "ignore:distutils Version classes are deprecated. Use packaging.version instead."
)
def test_dictify_xr_dataset(tmp_path: pathlib.Path) -> None:
    pytest.importorskip("netCDF4")

    # Define readonly dir
    readonly_dir = str(tmp_path / "readonly")
    fsspec.filesystem("file").mkdir(readonly_dir)
    config.set(cache_files_urlpath_readonly=readonly_dir)

    # Create sample dataset
    ds = xr.Dataset({"foo": [0]}, attrs={})
    with dask.config.set({"tokenize.ensure-deterministic": True}):
        token = dask.base.tokenize(ds)

    # Check dict
    actual = extra_encoders.dictify_xr_object(ds)
    print(fsspec.filesystem("file").ls(f"{tmp_path}/cache_files"))
    href = f"{readonly_dir}/{token}.nc"
    local_path = f"{tmp_path}/cache_files/{token}.nc"
    expected = {
        "type": "python_call",
        "callable": "cacholote.extra_encoders:decode_xr_dataset",
        "args": (
            {
                "type": "application/netcdf",
                "href": href,
                "file:checksum": f"{fsspec.filesystem('file').checksum(local_path):x}",
                "file:size": 669,
                "file:local_path": local_path,
            },
            {},
        ),
        "kwargs": {"chunks": {}},
    }
    assert actual == expected

    # Use href when local_path is missing or corrupted
    fsspec.filesystem("file").mv(local_path, href)
    xr.testing.assert_identical(ds, decode.loads(encode.dumps(actual)))


@pytest.mark.parametrize(
    "xarray_cache_type,ext,importorskip",
    [
        ("application/netcdf", ".nc", "netCDF4"),
        ("application/x-grib", ".grib", "cfgrib"),
        ("application/vnd+zarr", ".zarr", "zarr"),
    ],
)
@pytest.mark.parametrize("set_cache", ["file", "cads"], indirect=True)
@pytest.mark.filterwarnings(
    "ignore:GRIB write support is experimental, DO NOT RELY ON IT!"
)
@pytest.mark.filterwarnings(
    "ignore:distutils Version classes are deprecated. Use packaging.version instead."
)
def test_xr_cacheable(
    tmp_path: pathlib.Path,
    xarray_cache_type: str,
    ext: str,
    importorskip: str,
    set_cache: str,
) -> None:
    pytest.importorskip(importorskip)

    config.set(xarray_cache_type=xarray_cache_type)

    # cache-db to check
    con = config.get().engine.raw_connection()
    cur = con.cursor()

    expected = get_grib_ds()
    cfunc = cache.cacheable(get_grib_ds)

    for expected_counter in (1, 2):
        actual = cfunc()

        # Check hits
        cur.execute("SELECT counter FROM cache_entries", ())
        assert cur.fetchall() == [(expected_counter,)]

        # Check result
        if xarray_cache_type == "application/x-grib":
            xr.testing.assert_equal(actual, expected)
        else:
            xr.testing.assert_identical(actual, expected)

        # Check opened with dask (i.e., read from file)
        assert dict(actual.chunks) == {"longitude": (16,), "latitude": (31,)}


@pytest.mark.parametrize(
    "xarray_cache_type,ext,importorskip",
    [
        ("application/netcdf", ".nc", "netCDF4"),
        ("application/vnd+zarr", ".zarr", "zarr"),
    ],
)
def test_xr_corrupted_files(
    xarray_cache_type: str,
    ext: str,
    importorskip: str,
) -> None:
    pytest.importorskip(importorskip)
    import dask

    config.set(xarray_cache_type=xarray_cache_type)

    # Cache file
    fs, dirname = utils.get_cache_files_fs_dirname()
    expected = get_grib_ds()
    cfunc = cache.cacheable(get_grib_ds)
    cfunc()

    # Get cached file path
    with dask.config.set({"tokenize.ensure-deterministic": True}):
        root = dask.base.tokenize(expected)
    cached_path = f"{dirname}/{root}{ext}"
    assert fs.exists(cached_path)

    # Warn if file is corrupted
    fs.touch(cached_path, truncate=False)
    touched_info = fs.info(cached_path)
    with pytest.warns(UserWarning, match="checksum mismatch"):
        actual = cfunc()
    xr.testing.assert_identical(actual, expected)
    assert fs.info(cached_path) != touched_info

    # Warn if file is deleted
    fs.rm(cached_path, recursive=True)
    with pytest.warns(UserWarning, match="No such file or directory"):
        actual = cfunc()
    xr.testing.assert_identical(actual, expected)
    assert fs.exists(cached_path)


def test_xr_logging(log: pytest_structlog.StructuredLogCapture) -> None:
    config.set(logger=structlog.get_logger(), raise_all_encoding_errors=True)

    # Cache dataset
    cfunc = cache.cacheable(get_grib_ds)
    cached_ds = cfunc()
    urlpath = f"file://{cached_ds.encoding['source']}"
    tmpfile = log.events[0]["urlpath"]
    assert urlpath.rsplit("/", 1)[1] == tmpfile.rsplit("/", 1)[1]

    expected = [
        {
            "urlpath": tmpfile,
            "event": "start write tmp file",
            "level": "info",
        },
        {
            "urlpath": tmpfile,
            "write_tmp_file_time": log.events[1]["write_tmp_file_time"],
            "event": "end write tmp file",
            "level": "info",
        },
        {
            "urlpath": urlpath,
            "size": 22597,
            "event": "start upload",
            "level": "info",
        },
        {
            "urlpath": urlpath,
            "size": 22597,
            "upload_time": log.events[3]["upload_time"],
            "event": "end upload",
            "level": "info",
        },
        {
            "urlpath": urlpath,
            "event": "retrieve cache file",
            "level": "info",
        },
    ]
    assert log.events == expected


@pytest.mark.parametrize(
    "original_obj",
    (
        xr.DataArray([0], name="foo"),
        xr.DataArray([0], name="foo").to_dataset(),
    ),
)
def test_xr_roundtrip(original_obj: xr.Dataset | xr.DataArray) -> None:
    @cache.cacheable
    def cache_xr_obj(
        obj: xr.Dataset | xr.DataArray,
    ) -> xr.Dataset | xr.DataArray:
        return obj

    cached_obj = cache_xr_obj(original_obj)
    xr.testing.assert_identical(cached_obj, original_obj)
    assert original_obj.encoding.get("source") is None
    assert cached_obj.encoding.get("source") is not None



from __future__ import annotations

import pathlib

import fsspec

from cacholote import utils


def test_hexdigestify() -> None:
    text = "some random Unicode text \U0001f4a9"
    expected = "278a2cefeef9a3269f4ba1c41ad733a4"
    res = utils.hexdigestify(text)
    assert res == expected


def test_get_cache_files(tmp_path: pathlib.Path) -> None:
    assert utils.get_cache_files_fs_dirname() == (
        fsspec.filesystem("file"),
        str(tmp_path / "cache_files"),
    )


def test_copy_buffered_file(tmp_path: pathlib.Path) -> None:
    src = tmp_path / "test0"
    dst = tmp_path / "test1"
    with open(src, "wb") as f:
        f.write(b"test")
    with open(src, "rb") as f_src, open(dst, "wb") as f_dst:
        utils.copy_buffered_file(f_src, f_dst)
    assert open(src, "rb").read() == open(dst, "rb").read() == b"test"
